# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import io
import logging
import os
import sys

from byoc import stacks
from byot import fixtures
from byov import (
    config,
)
from byov.vms import (
    lxd,
)
from byov.tests import features

try:
    if sys.version_info < (3,):
        # novaclient doesn't support python3 (yet)
        from byov.vms import nova
except ImportError:
    pass
try:
    from byov.vms import ec2
    with_boto3 = bool(ec2)  # silly trick to pacify pyflakes
except ImportError:
    # No boto3, no ec2 vms.
    with_boto3 = False


# Useful shortcuts
patch = fixtures.patch
override_env = fixtures.override_env


def isolate_from_disk(test):
    """Provide an isolated disk-based environment.

    A $HOME directory is setup as well as an /etc/ one so tests can setup
    config files.
    """
    fixtures.set_uniq_cwd(test)
    fixtures.isolate_from_env(test)
    # Isolate tests from the user environment
    test.home_dir = os.path.join(test.uniq_dir, 'home')
    os.mkdir(test.home_dir)
    fixtures.override_env(test, 'HOME', test.home_dir)
    # Also isolate from the system environment
    test.etc_dir = os.path.join(test.uniq_dir, config.system_config_dir()[1:])
    os.makedirs(test.etc_dir)
    fixtures.patch(test, config, 'system_config_dir', lambda: test.etc_dir)


def set_uniq_vm_name(test, short=False):
    """To isolate tests from each other, created vms need a unique name.

    To keep those names legal and still user-readable we use the class name and
    the test method name. The process id is added so that the same test suite
    can be run on the same host in several processes.

    :param short: Defaults to False. Use only the test method and the pid. This
        helps with qemu unix socket names being limited in length.
    """
    if short:
        vm_name = '{}-{}'.format(test._testMethodName, os.getpid())
    else:
        vm_name = '{}-{}-{}'.format(test.__class__.__name__,
                                    test._testMethodName,
                                    os.getpid())
    # '_' are not valid in hostnames
    test.vm_name = vm_name.replace('_', '-')


# FIXME: Cruft is starting to accumulate here. 1) There is a tension around the
# isolation that happens below and contradicting needs to access user
# configuration (before isolation) /and/ setting up the test configuration
# (after isolation). 2) At least lxd and qemu backends co-exist rather than
# being selective -- vila 2019-11-16
def setup_tests_config(test):
    """Setup user provided configuration for tests.

    A self.uniq_dir/etc/byov.conf is created from
    ~/.config/byov/byov.conf-tests with a pre-defined set of config
    options.

    If files ending with '.conf-tests' exist under '~/.config/byov/conf.d',
    they are installed under 'test.etc_dir/conf.d' without the '-tests' suffix.

    """
    # Get a user configuration before isolation
    user_conf = config.VmStack(None)
    # qemu relies on images being provided (hence downloaded) by the user,
    # tests just reuse them.
    test.user_download_dir = user_conf.get('qemu.download.dir')
    # Make sure config stores don't leak via
    # byoc.stacks._shared_stores. vm.run_hook has to save all known config
    # stores before running to make sure any hook modification is properly
    # seen.
    fixtures.patch(test, stacks, '_shared_stores', {})
    features.test_requires(test, features.tests_config)
    # Share the lxd certificate. We could generate one on the fly but assume
    # instead that one is already available since we require the lxc client
    # already.
    lxd_conf_dir = os.path.expanduser('~/.config/lxc')
    isolate_from_disk(test)  # calls isolate_from_env
    fixtures.override_env(test, 'LXD_CONF', lxd_conf_dir)

    # By default, qemu tests start with a private images dir
    test.images_dir = os.path.join(test.uniq_dir, 'images')

    # Make user provided test config files visible to tests by installing them
    # under self.uniq_dir/etc/byov/conf.d
    def install(src, dst):
        with open(src) as s, open(dst, 'w') as d:
            d.write(s.read())
    install(features.tests_config.user_path,
            os.path.join(test.etc_dir, config.config_file_basename()))
    if features.tests_config.more_paths:
        confd_dir = os.path.join(test.etc_dir, 'conf.d')
        os.mkdir(confd_dir)
        for p in features.tests_config.more_paths:
            _, bname = os.path.split(p)
            # Drop the -tests suffix from the basename
            install(p, os.path.join(confd_dir, bname[:-len('-tests')]))
    # Create a config file for tests (all set() go to the section chosen at
    # Stack creation time)
    # FIXME: Ideally we'd want self.vm_name rather than None. Refactor tests to
    # allow that -- vila 2022-02-16
    conf = config.VmStack(None)
    # Some tests assumes -oLogLevel=ERROR is part of ssh.options to avoid the
    # 'Warning: Permanently added xxx to the list of known hosts.'
    test.assertTrue('-oLogLevel=ERROR' in conf.get('ssh.options'))
    # If there is no ssh key or it's pub counterpart is not in authorized_keys,
    # we won't be able to connect to vms
    key_path = conf.get('ssh.key')
    test.assertTrue(os.path.exists(key_path),
                    '{} does not exist'.format(key_path))
    authorized_keys = conf.get('ssh.authorized_keys')
    test.assertTrue(key_path + '.pub' in authorized_keys)
    # Some tests assumes byov-tests-<class>-debian exists (in the user env)
    # and that its digest matches the vm . The safest way to isolate this
    # properly is to just re-calculate the digest (in the test env).
    for kls_name, kls in (('lxd', lxd.Lxd),):
        vm_name = 'byov-tests-{}-debian'.format(kls_name)
        vm = kls(config.VmStack(vm_name))
        vm.econf.set('vm.name', vm_name)
        vm.econf.set('vm.class', kls_name)
        vm.econf.set('vm.setup.digest', vm.hash_setup())
        vm.econf.store.save_changes()
        vm.econf.store.unload()

    # ec2 test resources get some specific tags
    conf.set('ec2.instance.tags', 'test.id ' + test.id())
    conf.set('ec2.image.tags', 'test.id ' + test.id())
    conf.store.save_changes()
    conf.store.unload()


class _MyStreamHandler(logging.StreamHandler):
    """Work around an issue in python2 urllib3 library not using unicode.

    Can be deleted once we migrate to python3 novaclient.
    """

    def emit(self, record):
        msg = record.msg
        if sys.version_info < (3,) and isinstance(msg, str):
            record.msg = msg.decode('utf8')
        super(_MyStreamHandler, self).emit(record)


def override_logging(test, debug=None):
    """Setup a logging handler, restoring the actual handlers after the test.

    This assumes a logging setup where handlers are added to the root logger
    only.

    :param debug: When set to True, log output with level debug is sent to
        stdout.
    """
    # FIXME: Should we ask for '{logging.level}' instead ? -- vila 2022-01-07
    env_debug = bool(os.environ.get('DEBUG', False))
    if debug is None and env_debug:
        debug = env_debug
    if debug:
        stream = sys.stdout
        level = logging.DEBUG
        if with_boto3:
            # Silence boto a bit. without the following, roughly every http
            # request and response is traced. Because types and AMIs can't be
            # queried finely enough, it's mostly noise.
            for mod_name in ('boto3', 'botocore', 'urllib3.connectionpool'):
                logging.getLogger(mod_name).setLevel(logging.ERROR)
    else:
        stream = io.StringIO()
        level = logging.INFO
    root_logger = logging.getLogger(None)
    # Using reversed() below ensures we can modify root_logger.handlers as well
    # as providing the handlers in the right order for cleanups.
    for handler in reversed(root_logger.handlers):
        root_logger.removeHandler(handler)
        test.addCleanup(root_logger.addHandler, handler)
    # Install the new handler
    test.log_stream = stream
    new_handler = _MyStreamHandler(stream)
    test.addCleanup(root_logger.removeHandler, new_handler)
    root_logger.addHandler(new_handler)
    # Install the new level, restoring the actual one after the test
    test.addCleanup(root_logger.setLevel, root_logger.level)
    root_logger.setLevel(level)


def per_backend_release_arch_setup(test):
    """Prepare setting up vm for parametrization by backend, release and arch.

    This is used by the per_vm tests.
    """
    # Some classes require additional features
    required_features = {
        'docker': features.docker_client_feature,
        'ec2': features.ec2_creds,
        'lxd': features.lxd_client_feature,
        'nova': features.nova_creds,
        'qemu': features.qemu_img_feature,
        'scaleway': features.scaleway_creds,
    }
    if test.kls in required_features:
        f = required_features[test.kls]
        if not f.available():
            test.skipTest('{} is not available'.format(f.feature_name()))
    # Create a shared config
    conf = config.VmStack(test.vm_name)
    conf.set('vm.name', test.vm_name)
    conf.set('vm.class', test.kls)
    conf.set('vm.distribution', test.dist)
    conf.set('vm.release', test.series)
    conf.set('vm.architecture', test.arch)
    # Some classes require additional and specific setup

    def docker_setup(test, conf):
        conf.set('vm.published_as', '{vm.name}-public')
        image = conf.get('docker.image')
        with io.open('./Dockerfile', 'w', encoding='utf8') as f:
            f.write('''\
FROM {}
# FIXME: This is a mistake, building the image should not require running the
# container as ENTRYPOINT can interfere with it. -- vila 2022-03-01
# Force the container to stay up once started
ENTRYPOINT ["tail", "-f", "/dev/null"]
'''.format(image))

    def ec2_setup(test, conf):
        # Set to the empty string to use 'ec2.distribution.images'
        conf.set('ec2.image', '')
        conf.set('vm.published_as', '{vm.name}-public')
        if test.dist == "amazon":
            # at least on al2, dnf is not installed
            conf.set('dnf.command', 'sudo, yum, {dnf.options}')
            conf.set('dnf.options', '-y, -t')

    def lxd_setup(test, conf):
        conf.set('vm.published_as', '{vm.name}-public')

    def qemu_setup(test, conf):
        features.requires_existing_bridge(test, conf.get('qemu.bridge'))
        conf.set('qemu.networks',
                 '-net bridge,br={qemu.bridge}'
                 ' -net nic,macaddr={qemu.mac.address}')
        # Unless the test overrides it, qemu vms reuse downloads provided by
        # the user but use a private image dir.
        conf.set('qemu.download.dir', test.user_download_dir)
        conf.set('qemu.images.dir', test.images_dir)
        conf.set('vm.published_as', '{qemu.image}-public')
        # FIXME: qemu.image.setup and qemu.image.teardown ?
        # -- vila 2019-12-07

    def nova_setup(test, conf):
        image_id = nova.byov_image_name('cloudimg', test.series, test.arch)
        conf.set('nova.image', image_id)

    def scaleway_setup(test, conf):
        conf.set('scaleway.flavor', 'START1-XS')
        conf.set('scaleway.image',
                 '{vm.distribution}/{vm.release}/{vm.architecture}-xs')
        # conf.set('scaleway.flavor', 'C1')
        # conf.set('vm.architecture', 'armhf')
        features.test_requires(test, features.ScalewayImage(conf))

    specific_setups = {
        'docker': docker_setup,
        'ec2': ec2_setup,
        'lxd': lxd_setup,
        'nova': nova_setup,
        'qemu': qemu_setup,
        'scaleway': scaleway_setup,
    }
    if test.kls in specific_setups:
        setup = specific_setups[test.kls]
        setup(test, conf)
    # save all config changes
    conf.store.save()
    conf.store.unload()



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Setup a virtual machine from a config file.

Note: Most of the operations requires root access and this script uses ``sudo``
to get them.
"""
import errno
import hashlib
import logging
import os
import subprocess
import tempfile
import time

from byoc import stacks
from byov import (
    config,
    errors,
    logs,
    monitors,
    options,
    ssh,
    subprocesses,
    user_data,
)


logger = logging.getLogger(__name__)


class VM(object):
    """A virtual machine relying on cloud-init to customize installation."""

    setup_ip_timeouts = None  # Must be set by daughter classes
    setup_ssh_timeouts = None  # Must be set by daughter classes
    existing_vm_options = ['vm.name', 'vm.class', 'vm.distribution',
                           'vm.user', 'vm.user.home', 'vm.user.shell']

    def __init__(self, conf):
        # The conf where most of the options are defined
        self.conf = conf
        # The conf for the existing vm
        self.econf = config.ExistingVmStack(conf.get('vm.name'))
        # Seed files
        self._meta_data_path = None
        self._user_data_path = None
        self.ci_user_data = None

    def setup(self):
        raise NotImplementedError(self.setup)

    def teardown(self, force=False):
        self.cleanup_configs()

    def start(self):
        return self.run_start_hook()

    def stop(self):
        raise NotImplementedError(self.stop)

    def publish(self):
        raise NotImplementedError(self.publish)

    def unpublish(self):
        raise NotImplementedError(self.unpublish)

    # FIXME: logging missing -- vila 2020-02-17
    def pull(self, remote, local):
        """Pull a file from a vm.

        :param remote: A remote path inside the vm.

        :param local: A local path.
        """
        mode_bits = self._remote_mode_bits(remote)
        with open(local, 'w') as f:
            # We can't use 'scp' here as the ssh connection obeys ssh.options
            # and scp doesn't support all ssh options. So we just leverage the
            # ssh shell
            send_command = ['cat', remote]
            ssh_command = self.get_ssh_command(*send_command)
            send = subprocess.Popen(ssh_command,
                                    stdout=f,
                                    stderr=subprocess.PIPE)
            _, err = send.communicate()
            if send.returncode:
                raise errors.CommandError(ssh_command, send.returncode,
                                          '', err.decode('utf8'))
        os.chmod(local, mode_bits)
        return send.returncode

    # FIXME: logging missing -- vila 2020-02-17
    def push(self, local, remote):
        """Push a local file to a vm.

        :param local: A local path. If prefixed with '@' options are expanded
            before pushing.

        :param remote: A remote path inside the vm.
        """
        tmp = None
        if local.startswith('@'):
            src = local[1:]
            mode_bits = oct(self._mode_bits(local[1:]))[-3:]
            content = self.expand_file(local[1:])
            tmp = tempfile.NamedTemporaryFile(delete=False, mode='w')
            tmp.write(content)
            tmp.close()
            src = tmp.name
        else:
            src = local
            mode_bits = None
        try:
            ret = self.upload(src, remote, mode_bits)
        finally:
            if tmp is not None:
                os.remove(tmp.name)
        return ret

    # FIXME: logging missing -- vila 2020-02-17
    def save_existing_config(self):
        conf = self.conf
        for opt in self.existing_vm_options:
            # We need the raw values
            value = conf.get(opt, convert=False)
            if value:  # 'None' is never saved
                self.econf.set(opt, value)
        self.econf.store.save_changes()

    def cleanup_configs(self):
        for path in (self.console_path(), self.user_data_path(),
                     self.meta_data_path()):
            if os.path.exists(path):
                os.remove(path)
        config_dir_path = self.config_dir_path()
        # Delete the config directory if we can.
        try:
            os.rmdir(config_dir_path)
        except OSError as e:
            # Some server keys can still be there, keep them. If the dir is
            # already gone, don't complain.
            if e.errno not in (errno.ENOTEMPTY, errno.ENOENT):
                raise
        # Remove the vm from the existing ones
        name = self.conf.get('vm.name')
        try:
            del self.econf.store.sections[name]
        except KeyError:
            # The setup didn't went as far as creating the section
            pass
        # deleting a section is not supported by save_changes() so save() and
        # unload()
        self.econf.store.save()
        self.econf.store.unload()

    def hash_value(self, hasher, value):
        # No point in hashing none values, as soon as a value is set, it
        # introduces a difference.
        if value is not None:
            hasher.update(value.encode('utf8'))

    def hash_setup(self):
        """Returns a digest of the options used to setup the vm.

        This covers all relevant option values as well as the content of files
        referred to.

        This is a best effort that should cover most needs. Known limitations
        include: changes in files that are referred to indirectly, changes done
        inside the the vm after it was setup.
        """
        hasher = hashlib.md5()
        opts = []
        for more in ('vm.setup.digest.options', 'apt.setup.digest.options',
                     'pip.setup.digest.options',
                     'ssh.setup.digest_options'):
            l = self.conf.get(more)
            if l is not None:
                opts.extend(l)
        for opt in opts:
            odef = options.option_registry.get(opt)
            if isinstance(odef, options.PackageListOption):
                # The content of all the @file should be taken into account
                # (which PackageListOption conversion provides).
                value = self.conf.get(opt)
                if value is not None:
                    # We got a list, convert to a string
                    value = ' '.join(value)
            else:
                value = self.conf.get(opt, convert=False)
            self.hash_value(hasher, value)
        # FIXME: It would be nice to list the options below in a
        # 'vm.setup.digest.files' option but there are several blockers: 1) we
        # can't just dereference them as they may be None, 2) there is a mix of
        # scalar and lists (make them all lists ?) -- vila 2016-10-17
        setups = self.conf.get('vm.setup_scripts') or []
        uploads = self.conf.get('vm.uploaded_scripts') or []
        user_keys = self.conf.get('ssh.authorized_keys') or []
        server_keys = self.conf.get('ssh.server_keys') or []
        keys = user_keys + server_keys
        paths = [self.conf.get('vm.root_script'),
                 self.conf.get('vm.user_script')] + uploads + setups + keys
        for path in paths:
            try:
                with open(path) as f:
                    value = f.read()
            except Exception:
                # If an error has to be reported to the user, this is not from
                # here: the setup itself has/will fail and the digest will be
                # different, so just ignore
                continue
            self.hash_value(hasher, value)
        return hasher.hexdigest()

    # FIXME: could be named setup_over_shell -- vila 2022-01-19
    def setup_over_ssh(self):
        """Once shell access is setup, install  packages and run scripts."""
        # Save the defined config. Errors after that point should not break the
        # vm hard enough that one cannot shell into it and debug.
        self.save_existing_config()
        # Finish the setup
        self.update()
        self.install_packages()
        # Either update() or install_apt_packages() can trigger a reboot
        try:
            self.shell_captured(self.conf.get('vm.user.shell'),
                                cmd_input='test -f /var/run/reboot-required')
            logger.info('Rebooting {}...'.format(self.conf.get('vm.name')))
            self.stop()
            self.start()
        except errors.CommandError:
            # No reboot required, we're done
            pass
        self.run_setup_hook()
        self.run_setup_scripts()
        self.install_pip_packages()
        self.econf.set('vm.setup.digest', self.hash_setup())
        self.econf.store.save_changes()

    def shell(self, command, *args):
        if command and command.startswith('@'):
            ret, out, err = self.run_script(
                command[1:], args=args, captured=False)
        else:
            ssh_command = self.get_ssh_command(command, *args)
            ret, out, err = subprocesses.run(ssh_command)
        return ret, out, err

    def shell_captured(self, command, *args, **kwargs):
        cmd_input = kwargs.pop('cmd_input', None)
        # Compose Running [{cmd_input} under ]{command} on {vm.name}
        under = ''
        if cmd_input:
            under = ' ' + cmd_input + ' under '
        logger.debug('Running {}{} on {}'.format(
            under, command, self.conf.get('vm.name')))
        ssh_command = self.get_ssh_command(command, *args)
        retcode, out, err = subprocesses.run(ssh_command, cmd_input=cmd_input)
        logger.debug('Done with {}{}, it returned {} on {}'.format(
            command, under, retcode, self.conf.get('vm.name')))
        return retcode, out, err

    def update(self):
        if self.conf.get('vm.update'):
            logger.info('Updating {}...'.format(self.conf.get('vm.name')))
            pack_mgr = self.conf.expand_options(
                '{{vm.distribution}.package.manager}')
            if pack_mgr == 'apt':
                self.apt_get_update()
                self.apt_get_upgrade()
            elif pack_mgr == 'dnf':
                self.dnf_upgrade()

    def install_packages(self):
        packages = self.conf.get('vm.packages')
        if not packages:
            return 0
        logger.info('Installing packages {} on {}'.format(
            ' '.join(packages), self.conf.get('vm.name')))
        pack_mgr = self.conf.expand_options(
            '{{vm.distribution}.package.manager}')
        try:
            if pack_mgr == 'apt':
                self.do_apt_get(['install'] + packages)
            elif pack_mgr == 'dnf':
                self.do_dnf(['install'] + packages)
        except errors.CommandError:
            logger.debug('Failed to install packages', exc_info=True)
            raise

    def install_pip_packages(self):
        packages = self.conf.get('pip.packages')
        if not packages:
            return 0
        logger.info('Installing packages {} on {}'.format(
            ' '.join(packages), self.conf.get('vm.name')))
        install_command = ['install', '--user'] + packages
        try:
            self.do_pip(install_command)
        except errors.CommandError:
            logger.debug('Failed to install packages', exc_info=True)
            raise

    def run_hook(self, option):
        """Locally run a command defined by the user."""
        hook = self.conf.get(option)
        if not hook:
            return
        logger.info('Running {} on host for {}'.format(
            option, self.conf.get('vm.name')))

        def save_config_stores():
            # Hook can potentially change config files. Save all changes so
            # stores will be reloaded after each hook run.
            for store in stacks._shared_stores.values():
                store.save_changes()
                store.unload()

        tmp = None
        cmd = None
        if hook.startswith('@'):
            path = hook[1:]
            mode_bits = self._mode_bits(path)
            content = self.expand_file(path)
            logger.debug('Expanded script {}: {}'.format(path, content))
            tmp = tempfile.NamedTemporaryFile(delete=False, mode='w')
            tmp.write(content)
            tmp.close()
            os.chmod(tmp.name, mode_bits)
            cmd = [tmp.name]
        else:
            logger.debug('Hook: {}'.format(hook))
            cmd = ['/bin/sh', '-c', hook]
        try:
            save_config_stores()
            ret, out, err = subprocesses.run(cmd)
        except errors.CommandError:
            logger.error('Failed to run {}'.format(cmd), exc_info=True)
            raise
        finally:
            if tmp is not None:
                os.remove(tmp.name)
        return ret

    def run_setup_hook(self):
        """Locally run a setup command defined by the user."""
        return self.run_hook('vm.setup.hook')

    def run_start_hook(self):
        """Locally run a start command defined by the user."""
        return self.run_hook('vm.start.hook')

    def run_setup_scripts(self):
        scripts = self.conf.get('vm.setup_scripts')
        if scripts is None:
            return
        for path in scripts:
            try:
                self.run_script(path)
            except Exception:
                logger.debug('Failed to run {}'.format(path), exc_info=True)
                raise

    def _mode_bits(self, path):
        """Get mode bits for a path as string."""
        return os.stat(os.path.expanduser(path)).st_mode

    def _remote_mode_bits(self, path):
        """Get mode bits for a remote path as string."""
        _, out, _ = self.shell_captured('stat', '-c%a', path)
        return int(out, 8)

    def expand_file(self, local_path):
        with open(os.path.expanduser(local_path)) as f:
            content = f.read()
        expanded = self.conf.expand_options(content)
        return expanded

    def run_script(self, script_path, args=None, captured=True):
        if args is None:
            args = []
        logger.info('Running {}{} on {}'.format(
            script_path, ' '.join(args), self.conf.get('vm.name')))
        mode_bits = oct(self._mode_bits(script_path))[-3:]
        content = self.expand_file(script_path)
        tmp = tempfile.NamedTemporaryFile(delete=False, mode='w')
        tmp.write(content)
        tmp.close()
        # Get a temporary path (requires dropping the newline from the output)
        remote_path = self.shell_captured(
            'mktemp', '-t', 'byov.XXXXXX')[1][:-1]
        try:
            self.upload(tmp.name, remote_path, mode_bits=mode_bits)
            if captured:
                retcode, out, err = self.shell_captured(remote_path, *args)
                if retcode:
                    raise errors.CommandError([remote_path] + args,
                                              retcode, out, err)
            else:
                retcode, out, err = self.shell(remote_path, *args)
        finally:
            os.remove(tmp.name)
            self.shell_captured('rm', '-f', remote_path)
        logger.debug('Done with {}, it returned {} on {}'.format(
            script_path, retcode, self.conf.get('vm.name')))
        return retcode, out, err

    def upload(self, local_path, remote_path, mode_bits=None):
        if mode_bits is None:
            # Keep only the last 3 digits as that's what chmod will
            # accept. This will raise if path doesn't exist
            mode_bits = oct(self._mode_bits(local_path))[-3:]
        send_command = ['cat', local_path]
        send = subprocess.Popen(send_command,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        # We can't use 'scp' here as the ssh connection obeys ssh.options and
        # scp doesn't support all ssh options. So we just leverage the ssh
        # shell
        cat_command = ['cat', '-', '>', remote_path,
                       '&&', 'chmod', mode_bits, remote_path]
        ssh_command = self.get_ssh_command(*cat_command)
        recv = subprocess.Popen(ssh_command,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdin=send.stdout)
        # send.stdout.close()  # Allow 'recv' to get a SIGPIPE if 'send' exits
        out, err = recv.communicate()
        if recv.returncode:
            raise errors.CommandError(ssh_command, recv.returncode,
                                      out.decode('utf8'), err.decode('utf8'))
        out, err = send.communicate()
        if send.returncode:
            raise errors.CommandError(send_command, send.returncode,
                                      out.decode('utf8'), err.decode('utf8'))
        return send.returncode

    def dnf_upgrade(self):
        for timeout in self.conf.get('dnf.upgrade.timeouts'):
            try:
                self.do_dnf(['upgrade'])
                return  # We're done
            except errors.CommandError:
                msg = ('dnf upgrade failed,'
                       ' will sleep for {:.2f} seconds')
                logger.debug(msg.format(float(timeout)), exc_info=True)
                time.sleep(float(timeout))
        raise errors.ByovError('dnf upgrade never succeeded')

    def do_dnf(self, command):
        dnf_command = self.conf.get('dnf.command')
        # Filter empty values
        dnf_command = [arg for arg in dnf_command if arg]
        dnf_command.extend(command)
        return self.shell_captured(*dnf_command)

    def do_apt_get(self, command):
        apt_command = self.conf.get('apt.command')
        # Filter empty values
        apt_command = [arg for arg in apt_command if arg]
        apt_command.extend(command)
        return self.shell_captured(*apt_command)

    def do_pip(self, command):
        pip_command = self.conf.get('pip.command')
        # Filter empty values
        pip_command = [arg for arg in pip_command if arg]
        pip_command.extend(command)
        ret, out, err = self.shell_captured(
            self.conf.get('vm.user.shell'), cmd_input=' '.join(pip_command))
        return ret, out, err

    def apt_get_update(self):
        for timeout in self.conf.get('apt.update.timeouts'):
            try:
                self.do_apt_get(['update'])
                return  # We're done
            except errors.CommandError:
                msg = 'apt-get update failed, will sleep for {:.2f} seconds'
                logger.debug(msg.format(float(timeout)), exc_info=True)
                time.sleep(float(timeout))
        raise errors.ByovError('apt-get update never succeeded')

    def apt_get_upgrade(self):
        for timeout in self.conf.get('apt.upgrade.timeouts'):
            try:
                self.do_apt_get(['dist-upgrade'])
                return  # We're done
            except errors.CommandError:
                msg = ('apt-get dist-upgrade failed,'
                       ' will sleep for {:.2f} seconds')
                logger.debug(msg.format(float(timeout)), exc_info=True)
                time.sleep(float(timeout))
        raise errors.ByovError('apt-get dist-upgrade never succeeded')

    def ensure_dir(self, path):
        try:
            os.makedirs(path)
        except OSError as e:
            # python2 doesn't provide the exist_ok=True to makedirs
            if e.errno != errno.EEXIST:
                raise

    def remove(self, path):
        """Remove a path but don't complain if it's not there.

        Non-empty dirs still can't be deleted that way.

        Aka: It's Dead Jim.
        """
        if os.path.exists(path):
            os.remove(path)

    # FIXME: 'vm.config_dir' should be removed and fallouts addressed: calls to
    # self.ensure_dir(self.config_dir_path()) need to be removed too and
    # replaced by a better solution -- vila 2016-07-07
    def config_dir_path(self):
        return self.conf.get('vm.config_dir')

    def console_path(self):
        # The console is used to record 'setup' and 'start' outputs
        return os.path.join(self.config_dir_path(), 'console')

    def user_data_path(self):
        return os.path.join(self.config_dir_path(), 'user-data')

    def meta_data_path(self):
        return os.path.join(self.config_dir_path(), 'meta-data')

    def ssh_keygen(self, force=False):
        keys = self.conf.get('ssh.server_keys')
        if keys:
            for key in keys:
                key = os.path.abspath(os.path.expanduser(key))  # Just in case
                if force or not os.path.exists(key):
                    self.ensure_dir(os.path.dirname(key))
                    ssh.keygen(key, self.conf.get('vm.name'))

    def create_meta_data(self):
        self.ensure_dir(self.config_dir_path())
        self._meta_data_path = self.meta_data_path()
        with open(self._meta_data_path, 'w') as f:
            f.write(self.conf.get('vm.meta_data'))

    def create_user_data(self):
        ci_user_data = user_data.CIUserData(self.conf)
        ci_user_data.populate()
        self.ensure_dir(self.config_dir_path())
        self._user_data_path = self.user_data_path()
        with open(self._user_data_path, 'w') as f:
            f.write(ci_user_data.dump())
        self.ci_user_data = ci_user_data

    def capture_interface(self, line, iface_re):
        """Capture the required interface if it appears in 'line'.

        :param line: A line from the vm console.

        :param iface_re: The regexp matching a cloud-init output revealing an
            interface.

        :return: True if the interface was captured, False otherwise.
        """
        match = iface_re.match(line)
        if match is not None:
            iface, ip, mask, mac = match.groups()
            self.econf.set('vm.ip', ip)
            return True
        return False

    def scan_console_during_setup(self, console_size, console_path, cmd):
        """Scan the console output until the end of the setup.

        We add a specific part for cloud-init to ensure we properly detect
        the end of the run.

        :param console_size: The size of the console file before 'install' is
            run.

        :param console_path: The path to the console file.

        :param cmd: The setup command (used for error display).
        """
        console = monitors.TailMonitor(console_path, console_size)
        iface = '(eth0|ens3)'
        iface_re = logs.InterfaceRegexp(iface)
        try:
            for line in console.scan():
                # FIXME: We need some way to activate this dynamically (conf
                # var defaulting to environment variable OR cmdline parameter ?
                # -- vila 2013-02-11
                # print('read: [{}]'.format(line))  # so useful for debug...
                self.capture_interface(line, iface_re)
        except (errors.ConsoleEOFError, errors.CloudInitError):
            # FIXME: No explicit test covers this path -- vila 2013-02-15
            # FIXME: errors.ConsoleEOFError says 'Suspicious line', that's
            # confusing -- vila 2019-10-16
            err_lines = ['Suspicious output line from cloud-init.\n',
                         '\t' + console.lines[-1],
                         'Check the configuration:\n']
            with open(self._meta_data_path) as f:
                err_lines.append('meta-data content:\n')
                err_lines.extend(f.readlines())
            with open(self._user_data_path) as f:
                err_lines.append('user-data content:\n')
                err_lines.extend(f.readlines())
            raise errors.CommandError(cmd, console.proc.returncode,
                                      '\n'.join(console.lines),
                                      ''.join(err_lines))

    def scan_console_during_start(self, console_size, cmd):
        """Scan the console output while the instance starts.

        This is used to capture the network addresses displayed by cloud-init.

        :param console_size: The size of the console file before 'start' is
            run.

        :param cmd: The start command (used for error display).
        """
        console = monitors.TailMonitor(self.console_path(), console_size)
        iface = '(eth0|ens3)'
        iface_re = logs.InterfaceRegexp(iface)
        try:
            for line in console.scan():
                if self.capture_interface(line, iface_re):
                    # We're done, no need to scan anymore (and work around the
                    # fact that 'scan' will otherwise terminate when the vm is
                    # stopped ;)
                    # FIXME: This fails to exit the loop when no IP can be
                    # provided (seen while investigating
                    # https://bugs.launchpad.net/bugs/1465196)
                    # -- vila 2015-11-01
                    return
        except errors.ConsoleEOFError:
            # We're done
            pass

    def get_ssh_user_host_port(self):
        c = self.conf
        return (c.get('ssh.user'), c.get('ssh.host'), c.get('ssh.port'))

    def get_ssh_command(self, command, *args):
        cmd = ['ssh']
        user, host, port = self.get_ssh_user_host_port()
        options = self.conf.get('ssh.options')
        if options:
            cmd += options
        cmd += ['{}@{}'.format(user, host)]
        if port:
            cmd += ['-p', port]
        if command is not None:
            cmd += [command]
            if args:
                cmd += args
        return cmd

    def discover_ip(self):
        raise NotImplementedError(self.discover_ip)

    def wait_for_ip(self):
        logger.info('Waiting for an IP address for {}...'.format(
            self.conf.get('vm.name')))
        ip = None
        stos = self.conf.get(self.setup_ip_timeouts)
        me = self.wait_for_ip.__name__
        for attempt, sleep in enumerate(stos):
            try:
                if attempt > 1:
                    logger.debug('Re-trying {} {}/{}'.format(
                        me, attempt, stos.retries))
                ip = self.discover_ip()
                self.econf.set('vm.ip', ip)
                logger.info('{} IP is {}'.format(self.conf.get('vm.name'),
                                                 self.conf.get('vm.ip')))
                return
            except Exception:
                logger.debug(
                    'IP not yet available for {}'.format(
                        self.conf.get('vm.name')),
                    exc_info=True)
            # FIXME: metric  -- vila 2015-06-25
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, stos.retries))
            time.sleep(sleep)
        raise errors.ByovError('{name} never received an IP',
                               name=self.conf.get('vm.name'))

    def wait_for_ssh(self):
        logger.info('Waiting for {} ssh...'.format(self.conf.get('vm.name')))
        stos = self.conf.get(self.setup_ssh_timeouts)
        me = self.wait_for_ssh.__name__
        exc = None
        for attempt, sleep in enumerate(stos, start=1):
            try:
                if attempt > 1:
                    logger.debug('Re-trying {} {}/{}'.format(
                        me, attempt, stos.retries))
                ret, out, err = self.shell_captured('whoami')
                # Success, clear the exception
                exc = None
                break
            except errors.CommandError as e:
                logger.debug('ssh is not up yet', exc_info=True)
                exc = e
                # FIXME: metric  -- vila 2015-06-25
                pass
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, stos.retries))
            time.sleep(sleep)
        if exc is not None:
            # Re-raise the last seen exception
            raise exc

    def _wait_for_cloud_init(self, timeouts_name):
        logger.info('Waiting for cloud-init on {}...'.format(
            self.conf.get('vm.name')))
        # Now let's track cloud-init completion
        # We know ssh is up and running
        citos = self.conf.get(timeouts_name)
        me = self.wait_for_cloud_init.__name__
        exc = None
        # boot-finished is deleted when cloud-init starts (which is before the
        # guest get network access) and created once it finishes. Since we use
        # ssh to check, there is no race around the deletion so it's always
        # safe to wait for the file to appear.
        for attempt, sleep in enumerate(citos, start=1):
            try:
                if attempt > 1:
                    logger.debug('Re-trying {} {}/{}'.format(
                        me, attempt, citos.retries))
                self.shell_captured(
                    self.conf.get('vm.user.shell'),
                    cmd_input='test -f /var/lib/cloud/instance/boot-finished')
                # Success, clear the exception in case it was set
                exc = None
                break
            except errors.CommandError as e:
                logger.debug('cloud-init is not done yet', exc_info=True)
                exc = e
                # FIXME: logging + metric  -- vila 2015-06-25
            logger.debug('Sleeping {:.2f} seconds for {} {}/{}'.format(
                sleep, me, attempt, citos.retries))
            time.sleep(sleep)
        # FIXME: Worth downloading cloud-init[-output].log locally. It's the
        # first place to look when debugging cloud-init issues
        # -- vila 2019-12-10
        if exc is not None:
            # Re-raise the last seen exception
            raise exc
        # FIXME: We need to check if cloud-init failed or not but
        # /var/log/cloud-init[-output].log contain all executions :-/ So
        # checking only for the last execution is not trivial
        # -- vila 2016-05-13



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014, 2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import os
import logging
import unittest

from byot import assertions

from byov.tests import fixtures


class TestIsolateFromDisk(unittest.TestCase):

    def test_disk_preserved(self):
        real_home = os.path.expandvars('$HOME')

        class Inner(unittest.TestCase):

            def test_overridden(self):
                fixtures.isolate_from_disk(self)
                # We know expanduser will use $HOME which is set by
                # isolate_from_disk
                self.assertNotEqual(real_home, os.path.expandvars('$HOME'))
                path = os.path.expanduser('~/testing')
                with open(path, 'w') as f:
                    f.write('whatever')
                # Make sure the file is created in the right place
                self.assertTrue(os.path.exists(path))

        assertions.assertSuccessfullTest(self, Inner('test_overridden'))
        # Make sure the file wasn't created in the wrong place
        self.assertFalse(os.path.exists('~/testing'))


class TestOverrideLogging(unittest.TestCase):

    def test_handlers_preserved(self):
        # Keep a copy of the existing handlers
        installed_handlers = [h for h in logging.getLogger().handlers]

        class Inner(unittest.TestCase):

            def test_overridden(self):
                fixtures.override_logging(self, debug=True)
                root_logger = logging.getLogger()
                assertions.assertLength(self, 1, root_logger.handlers)
                # Careful here, debug=True means any log goes to stdout (aka
                # leaks). Alternatively, aka the day it fails, this should be
                # rewritten by redirecting stdout
                self.assertEqual(self.log_stream,
                                 root_logger.handlers[0].stream)
                self.assertEqual(logging.DEBUG, root_logger.level)

        assertions.assertSuccessfullTest(self, Inner('test_overridden'))
        self.assertEqual(installed_handlers, logging.getLogger().handlers)
